// gDirRec.cpp -- Version 0.7

#include "gdir.h"
#include "gDirRec.h"
////////////////////////////////////////////////////////////
t_uint64 gdr_calc_block_size (t_uint64 v, sDuOpt& duOpt)
{
 // 'stat' will give number of blocks in 512octet units
 t_uint64 result = v / 2 + (v==0 || (v%2)!=0);
 return result;
}

t_uint64 gdr_calc_byte_size (t_uint64 v, t_uint64 blocksInFile, sDuOpt& duOpt)
{
 t_uint64 result;
 if ( duOpt.doExact ) return v;
#ifdef gDOS_SPEC
 if ( blocksInFile==0 ) {
     // stat returned a poor result (like with DevC++ default libs): assuming 4096-octet blocks
     blocksInFile = (v / (t_uint64)4096) + ((v % 4096)!=0 || v==0);
     return blocksInFile*4096;
 }
#endif //gDOS_SPEC
 result = blocksInFile*512;
 return result;
}
////////////////////////////////////////////////////////////
int gdr_du_dir (char* str,
		FILE* fOut,
		FILE* fRepErr,
		gHashTriple& myHash,
		sDuOpt& duOpt,
		int verboseLevel,
		t_uint64& byteSize,
		t_uint64& blockSize,
		int& recControl,
		int& nErrors)
{
 int thisError;
 unsigned k, n;
 bool isDir;
 gKey hKey( gKey::e_UInt64 );

 ASSERTION(fRepErr,"fRepErr");
 if ( duOpt.aMaxDepth>-1 && recControl>=duOpt.aMaxDepth ) {
     if ( verboseLevel>=3 ) {
	 fprintf(fRepErr,"Depth level, skipping: %s\n",str);
     }
     return 0;
 }

 if ( duOpt.doExact ) {
     blockSize = 0;
     byteSize = 0;
 }
 else {
     blockSize = 1;  // One dir occupies only one minor(512b) block
     byteSize = gdr_calc_byte_size( 0, blockSize, duOpt );
 }

 gDir aDir( str );
 thisError = aDir.lastOpError;
 if ( thisError!=0 ) {
     fprintf(fRepErr,"Cannot access directory (%s): %s\n",
	     gFileControl::Self().ErrorStr( thisError ),
	     str);
     nErrors++;
     return 2;
 }

 if ( recControl>9999 ) {
     fprintf(fRepErr,"Too many nested dirs... (at least %d)\n",recControl);
     return -1;
 }
 recControl++;


 // error will count number of errors
 n = aDir.N();
 for (k=1; k<=n; k++) {
     bool isHardLink;
     t_inode inode;
     t_uint16 nLinks;
     char* aName;
     gFileSysName* fSysPtr = (gFileSysName*)aDir.GetObjectPtr( k );
     gFileStat* pStat = fSysPtr->GetStatPtr();
     aName = fSysPtr->Str();
     gString sName( str );
     sName.Add( aName );
     aName = sName.Str();
     inode = pStat->status.inode;
     nLinks = pStat->status.nLinks;
     isDir = fSysPtr->IsDirectory();

     // Check if the hard-link was already counted.
     // A stupid question: "which is the master file when having 'a' and 'b'?"
     // Answer is: the first that was created will appear in the count...
     if ( (isHardLink=(isDir ? nLinks>2 : (nLinks>1)))==true ) {
	 hKey.Set( inode );
	 if ( duOpt.doCountHardLinks==false ) {
	     // ...then it's a hard-link and we do not count it
	     if ( myHash.FindKey( hKey ) ) continue;
	     myHash.AddTriple( hKey, aName, nil );
	 }
	 else {
	     if ( verboseLevel>=3 && duOpt.doCountHardLinks==true ) myHash.AddTriple( hKey, aName, nil );
	 }
     }

     t_uint64 aByteSize=0;
     t_uint64 aBlockSize=0;
     if ( isDir ) {
	 if ( verboseLevel>=3 ) fprintf(fRepErr,"Recursing (%d) into '%s'\n",recControl,aName);
	 // Recursive call to 'gdr_du_dir'
	 thisError = gdr_du_dir( aName, fOut, fRepErr, myHash, duOpt, verboseLevel, aByteSize, aBlockSize, recControl, nErrors );
	 if ( thisError!=0 ) {
	     nErrors++;
	     continue;
	 }
	 gdr_du_print( aName, fOut, *pStat, duOpt, verboseLevel, aByteSize, aBlockSize, isHardLink, false );
     }
     else {
	 t_uint32 blocksInFile;
	 aBlockSize = (t_uint64)(blocksInFile = pStat->statusL.blocks);
	 if ( pStat->IsLink() )
	     aByteSize = 0;
	 else
	     aByteSize = gdr_calc_byte_size( (t_uint64)(pStat->statusL.size), (t_uint64)blocksInFile, duOpt );
/*	 if ( duOpt.DoCheck() || duOpt.doShowAll ) {
	     do_du_print( aName, fOut, *pStat, duOpt, aByteSize, aBlockSize, isHardLink, false );
	 }
*/
	 if ( duOpt.doShowAll ) {
	     gdr_du_print( aName, fOut, *pStat, duOpt, verboseLevel, aByteSize, aBlockSize, isHardLink, false );
	 }
     }
     blockSize += aBlockSize;
     byteSize += aByteSize;
 }//end FOR k

 recControl--;
 return 0;
}
////////////////////////////////////////////////////////////
int gdr_du_print (char* str,
		  FILE* fOut,
		  gFileStat& fStat,
		  sDuOpt& duOpt,
		  int verboseLevel,
		  t_uint64 byteSize,
		  t_uint64 blockSize,
		  bool isHardLink,
		  bool isArg)
{
 char fmtStr[16];
 t_uint64 uVal64;
 unsigned long long val;
 bool isVeryVerbose( verboseLevel>=3 );
 bool doShowHardLinkSymbol = isVeryVerbose && isHardLink;

 if ( duOpt.doSummarize ) {
     if ( isArg==false ) return 0;
 }
 if ( duOpt.doByByteSize==false && duOpt.doByBlockSize==false ) {
     fprintf(fOut,"%s\n",str);
     return 0;
 }
 sprintf(fmtStr,duOpt.doSupportLL?"%s%%%s%ullu\t%%s\n":"%s%%%s%ulu\t%%s\n",
	 doShowHardLinkSymbol?"@":"\0",
	 duOpt.sFmtBasicDigits.Str(),
	 duOpt.GetMaxDigits()-(unsigned)doShowHardLinkSymbol);
 uVal64 = byteSize;
 if ( duOpt.doByBlockSize ) {
#ifdef gDOS_SPEC
     // DevC++ does not support 'blockSize' properly:
     // just calc. the most probable number of blocks used (assuming adjacent disk blocks!)
     uVal64 = (byteSize / 1024) + (byteSize==0 || (byteSize%1024)!=0);
#else
     uVal64 = gdr_calc_block_size( blockSize, duOpt );
#endif //gDOS_SPEC
 }
 val = (unsigned long long)uVal64;
 if ( duOpt.doSupportLL )
	fprintf(fOut,fmtStr,val,str);
 else
	fprintf(fOut,fmtStr,(unsigned long)val,str);

 // e.g.: printf("%-8llu%s\n",val,str) : DevC++ does not support printf long-long
 return 1;
}
////////////////////////////////////////////////////////////
int gdr_du_dir_list (char* str,
		     FILE* fOut,
		     FILE* fRepErr,
		     gHashTriple& myHash,
		     sDuOpt& duOpt,
		     int verboseLevel,
		     gList& lPaths,
		     t_uint64& byteSize,
		     t_uint64& blockSize,
		     int& recControl,
		     int& nErrors)
{
 // Sorry for almost copying 'gdr_du_dir' into here,
 // but I was terribly lazy, and 'gdr_du_dir' worked so well!

 int thisError;
 unsigned k, n;
 bool isDir;
 gKey hKey( gKey::e_UInt64 );

 ASSERTION(fRepErr,"fRepErr");
 if ( duOpt.aMaxDepth>-1 && recControl>=duOpt.aMaxDepth ) {
     if ( verboseLevel>=3 ) {
	 fprintf(fRepErr,"Depth level, skipping: %s\n",str);
     }
     return 0;
 }

 gDir aDir( str );
 thisError = aDir.lastOpError;
 if ( thisError!=0 ) {
     fprintf(fRepErr,"Cannot access directory (%s): %s\n",
	     gFileControl::Self().ErrorStr( thisError ),
	     str);
     nErrors++;
     return 2;
 }

 if ( recControl>9999 ) {
     fprintf(fRepErr,"Too many nested dirs... (at least %d)\n",recControl);
     return -1;
 }
 recControl++;


 // error will count number of errors
 n = aDir.N();
 for (k=1; k<=n; k++) {
     bool isHardLink;
     t_inode inode;
     t_uint16 nLinks;
     char* aName;
     gFileSysName* fSysPtr = (gFileSysName*)aDir.GetObjectPtr( k );
     gFileStat* pStat = fSysPtr->GetStatPtr();
     aName = fSysPtr->Str();
     gString sName( str );
     sName.Add( aName );
     aName = sName.Str();
     inode = pStat->status.inode;
     nLinks = pStat->status.nLinks;
     isDir = fSysPtr->IsDirectory();

     // Check if the hard-link was already counted.
     // A stupid question: "which is the master file when having 'a' and 'b'?"
     // Answer is: the first that was created will appear in the count...
     if ( (isHardLink=(isDir ? nLinks>2 : (nLinks>1)))==true ) {
	 hKey.Set( inode );
	 if ( duOpt.doCountHardLinks==false ) {
	     // ...then it's a hard-link and we do not count it
	     if ( myHash.FindKey( hKey ) ) continue;
	     myHash.AddTriple( hKey, aName, nil );
	 }
	 else {
	     if ( verboseLevel>=3 && duOpt.doCountHardLinks==true ) myHash.AddTriple( hKey, aName, nil );
	 }
     }

     t_uint64 aByteSize=0;
     t_uint64 aBlockSize=0;
     if ( isDir ) {
	 if ( verboseLevel>=3 ) fprintf(fRepErr,"Recursing (%d) into '%s'\n",recControl,aName);
	 // Recursive call to 'gdr_du_dir<...>'
	 thisError = gdr_du_dir_list( aName, fOut, fRepErr, myHash, duOpt, verboseLevel, lPaths, aByteSize, aBlockSize, recControl, nErrors );
	 if ( thisError!=0 ) {
	     nErrors++;
	     continue;
	 }
	 lPaths.Add( aName );
     }
     else {
	 if ( pStat->IsLink()==false ) {
	     lPaths.Add( aName );
	     DBGPRINT_MIN("DBG: ADDED: %s\n",aName);
	 }
     }
 }//end FOR k

 recControl--;
 return 0;
}
////////////////////////////////////////////////////////////

