// gNormIoHelp.cpp -- Version 0.6

#include <unistd.h>
#include <errno.h>

#include "gfile.h"
#include "gNormIoHelp.h"
#include "gNormStubs.h"
////////////////////////////////////////////////////////////
char* gio_getcwd (char* buf, size_t size)
{
 char* str;
 ASSERTION(buf,"gio_getcwd (1)");
 ASSERTION(size,"gio_getcwd (2)");
 str = getcwd( buf, size );
 gStorageControl::Self().lastOpError = errno;
 return str;
}
////////////////////////////////////////////////////////////
off_t gio_file_seek (int fildes, off_t offset, int whence)
{
 if ( fildes<0 ) return -1;
 off_t seeked( gns_lseek( fildes, offset, whence ) );
#ifdef DEBUG_MIN
 if ( whence==SEEK_SET ) return seeked;
 printf("DBG: enter seek value%s: %d: ",whence==SEEK_END?" (SEEK_END)":"\0",seeked);
 seeked=-2;
 while ( fscanf( stdin, "%ld", &seeked )!=1 && seeked<-1 ) ;
#endif //DEBUG_MIN
 return seeked;
}

int gio_stable_file (int fd, FILE* fRepErr, int doProbe, off_t& seeked)
{
 int retry=1, maxRetries=10;
 off_t seekedCheck;
 ssize_t did_read;
 char blindBuf[ 128 ];
 unsigned secSleep;

 ASSERTION(fd>=0,"fd>=0");
 seeked = gio_file_seek( fd, 0, SEEK_END );
 if ( seeked<0 ) return 1;

 if ( gio_file_seek( fd, 0, SEEK_SET )<0 ) return 1;

 for (secSleep=5;
      retry<=maxRetries && (seekedCheck = gio_file_seek( fd, 0L, SEEK_END ))!=seeked;
      retry++, secSleep+=10) {
     if ( gio_file_seek( fd, seeked, SEEK_SET )!=seeked ) {
	 if ( fRepErr )
	     fprintf(fRepErr,"File shrunk (below %lu bytes)?\n",seeked);
	 return 4;  // Severe error, flie shrunk?
     }
     if ( doProbe==1 ) {
	 if ( fRepErr )
	     fprintf(fRepErr,"File size differs, retry (%d/%d), sleeping %u seconds\n",
		     retry,
		     maxRetries,
		     secSleep);
	 gFileControl::Self().SecSleep( (t_uint32)secSleep );
     }
 }

 did_read = read( fd, blindBuf, sizeof(blindBuf) );
 if ( did_read>0 ) return 3;  // Unable to know end-point of file

 // Rewind!
 if ( gio_file_seek( fd, 0, SEEK_SET )<0 ) return 1;

 return 0;
}
////////////////////////////////////////////////////////////

