// gStrings.cpp -- Version 0.1

#define thisProgramVersion "Version 0.1"
#define thisProgramCopyright "Prized Season & Sons"
#define thisProgramYear 2007
#define thisProgramCopy "This is free software (GPL)\n\
There is no warranty, not even for MERCHANTABILITY or\n\
FITNESS FOR A PARTICULAR PURPOSE."

#include <stdio.h>
#include <string.h>
#include "garg.h"

#include "gSupStrUni.h"

#ifdef gDOS_SPEC
#include <windows.h>
#endif //gDOS_SPEC

// Global (uggly, not using libgobjeio...)
gOptUnCode optUncode;
sUnCodeBin ucBin;
////////////////////////////////////////////////////////////
struct sOptStrgs {
    sOptStrgs ()
	: isVerbose( false ),
	  isVeryVerbose( false ),
	  isOutDOS( false ),
	  isDumpLinear( false ),
	  doSkip8Bit( false ),
	  doViewAll( false ),
	  caseKind( 0 ) {
    }
    ~sOptStrgs () {
    }

    bool isVerbose, isVeryVerbose;
    bool isOutDOS;
    bool isDumpLinear;
    bool doSkip8Bit;
    bool doViewAll;

    int caseKind;	// 0:none; 1:up; 2:down/lower-case

    gString sDumpFile;
    gString sCase;
};
////////////////////////////////////////////////////////////
int print_help (char* progStr)
{
 char
     *msgHelp = "%s - %s\n\n\
Usage:\n\
        %s command [options] [file ...]\n\
\n\
Options are:\n\
   -h              This help (or --help / --version)\n\
   -v              Verbose (use twice, more verbose)\n\
   -8              Skip 8bit ISO 8859-1 latin chars (or --skip-8bit)\n\
   -c S            Case S (or --case S)\n\
   -o X            Output to file X (or --output)\n\
   --dos-mode      DOS-mode (CR&LF)\n\
   --dump-linear   Linear dump\n\
   --view          View printable chars\n\
\n\
Examples:\n\
	gstrings -c up mytext.txt\n\
\n";

 printf(msgHelp,
	progStr,
	thisProgramVersion,
	progStr);
 return 0;
}
////////////////////////////////////////////////////////////
int print_version (char* progStr)
{
 char
     *msgVersion = "%s - %s\n\
Written by Henrique Moreira.\n\
\n\
Copyright (C) %u %s.\n\
%s\n";

 printf(msgVersion,
	progStr,
	thisProgramVersion,
	thisProgramYear,thisProgramCopyright,
	thisProgramCopy);
 return 0;
}
////////////////////////////////////////////////////////////
// glib required functions
////////////////////////////////////////////////////////////
int gglobal_EventTerminate (int v)
{
 return 0;
}
////////////////////////////////////////////////////////////
// Specific functions
////////////////////////////////////////////////////////////
int go_file (FILE* fIn, FILE* fOut, FILE* fRepErr,
	     sUnCodeBin& uc, gOptUnCode& unCode, sOptStrgs& opt)
{
 int error;
 unsigned maxChr( uc.n ), unChr;
 t_uchar chr( '\0' );
 t_uchar refChrUse( '\0' );
 t_uchar lastChr( '\0' );
 bool lastIsPrintable( false );

 char twoLetterC2nd;
 t_int16 refCode;
 t_gUniPage ucCompat;
 t_gUniPage ucAlt;
 int letterUp1Down2;

 ASSERTION(fIn,"fIn");
 ASSERTION(fOut,"fOut");

 DBGPRINT("DBG: go_file maxChr=%u, doViewAll?%c\n",
	  maxChr,
	  ISyORn(opt.doViewAll));

 if ( maxChr<256 ) return 8;

 for ( ; (error = fscanf(fIn,"%c",&chr)!=1)==0; ) {
     unChr = (unsigned)chr;
     bool isPrintable( (uc.mapTab[ unChr ].extUtMask & EXTUMASK_ISPRINT)!=0 );
     if ( opt.doSkip8Bit==true && unChr>=127 ) {
	 lastIsPrintable = false;
	 continue;
     }
     refCode = unCode.refCode[ unChr ];
     refChrUse = unCode.refChrUse[ unChr ];
     ASSERTION(refCode>=0,"refCode>=0");

     if ( opt.doViewAll ) {
	 gString* name = uc.mapTab[ unChr ].symbolName;
	 char* symNameStr = name ? name->Str() : nil;
	 // View all printable chars
	 //if ( unChr > 128 ) {
	 //    printf("\nDBG: %ud '%c' isPrintable?%c\n",
	 //	    unChr,
	 //	    chr,
	 //	    ISyORn(isPrintable));
	 //}
	 if ( isPrintable ) {
	     fprintf(fOut,"%c",chr);
	 }
	 else {
	     fprintf(fOut,"[[%s]]\n",symNameStr);
	 }
	 continue;
     }

     if ( refCode<=0 ) {
	 lastIsPrintable = false;
	 continue;
     }

     ucCompat = uc.mapTab[ unChr ].ucCompat;

     twoLetterC2nd = uc.mapTab[ unChr ].twoLetterCat[ 1 ];
     if ( uc.mapTab[ unChr ].twoLetterCat[ 0 ]=='L' )
	 letterUp1Down2 = (twoLetterC2nd=='u') + ((twoLetterC2nd=='l')*2);
     else
	 letterUp1Down2 = 0;

     switch ( opt.caseKind ) {
     case 0:
	 if ( refChrUse==0 ) {
	     lastIsPrintable = false;
	     continue;
	 }
	 chr = refChrUse;
	 break;

     case 1:
	 ucAlt = uc.mapTab[ unChr ].ucUpper;
	 if ( ucAlt )
	     chr = (char)ucAlt;
	 else
	     if ( letterUp1Down2==2 )
		 chr = '?';
	 break;

     case 2:
	 ucAlt = uc.mapTab[ unChr ].ucDown;
	 if ( ucAlt )
	     chr = (char)ucAlt;
	 break;

     case 8:
	 chr = (t_uchar)refCode;
	 // Apparently similar to:
	 //	char chrCompat( ucCompat );
	 //	chr = chrCompat ? chrCompat : refChrUse;
	 break;

     default:
	 ASSERTION_FALSE("Uops");
     }

     bool lastPrintOrNew( lastIsPrintable==true || lastChr=='\n' );
     fprintf(fOut,"%s%c",
	     lastPrintOrNew ? "\0" : "\n",
	     chr);
     lastIsPrintable = true;
     lastChr = chr;
 }

 return 0;
}
////////////////////////////////////////////////////////////
int go_cmd (gArg& arg, FILE* fOut, FILE* fRepErr, sOptStrgs& opt, int aCmd)
{
 int error = 0, thisError, countIter = 0;
 char* str;
 unsigned i, n = arg.N();
 FILE* fIn = nil;

 ASSERTION(fRepErr,"fRepErr");
 ASSERTION(fOut,"fOut");

 for (i=1; i<=n; i++) {
     str = arg.Str( i );
     if ( str[0]==0 ) continue;
     if ( opt.isVeryVerbose || (opt.isVerbose==true && n>1) ) fprintf(stderr,":::%s\n",str);
     countIter++;
     fIn = fopen( str, "rb" );
     thisError = fIn ? 0 : 2;
     if ( thisError==0 ) {
	 thisError = go_file( fIn, fOut, fRepErr,
			      ucBin, optUncode, opt );
	 fclose( fIn );
     }
     if ( thisError!=0 ) {
	 fprintf(fRepErr,"Error (%d) handling: %s\n",thisError,str);
	 if ( error==0 ) error = thisError;
     }
     DBGPRINT("DBG: 'go_file' returned %d count=%d\n",thisError,countIter);
 }

 if ( countIter==0 ) {
     error = go_file( stdin, fOut, fRepErr, ucBin, optUncode, opt );
 }

 return error;
}
////////////////////////////////////////////////////////////
int go (char** argv, char** envp, int& result)
{
 int error, thisError = -1;
 gArg arg( argv, envp );
 sOptStrgs opt;

 unsigned n;
 FILE* fOut = stdout;
 FILE* fRepErr = stderr;

 result = 0;

 arg.AddParams( 1, "-vv|--verbose -h|--help --version\
 -c:%s|--case:%s\
 -o:%s|--output:%s\
 -8|--skip-8bit\
 --dos-mode\
 --dump-linear\
 --view\
" );

 error = arg.FlushParams();
 n = arg.N();
 arg.FindOptionOccurr( "verbose", opt.isVerbose, opt.isVeryVerbose );

 if ( arg.FindOption('h') || error!=0 ) {
     print_help( arg.Program() );
     return 0;
 }
 if ( arg.FindOption("version") ) {
     print_version( arg.Program() );
     return 0;
 }

 opt.isOutDOS = arg.FindOption("dos-mode");
 opt.isDumpLinear = arg.FindOption("dump-linear");
 opt.doSkip8Bit = arg.FindOption('8');
 opt.doViewAll = arg.FindOption("view");

 if ( arg.FindOption('c',opt.sCase) && opt.sCase.IsEmpty()==false ) {
     gString s( opt.sCase );
     int caseKind( 0 );
     caseKind += s.Match( "up", true ) || s.Match( "upper", true );
     caseKind += (s.Match( "down", true ) || s.Match( "lower", true ))*2;
     caseKind += s.Match( "normal", true )*8;
     if ( caseKind==0 ) {
	 fprintf(stderr,"Invalid case specified: either normal, 'up'/'upper' or 'down'/'lower' can be chosen.\n");
	 return 1;
     }
     opt.caseKind = caseKind;
 }

 if ( arg.FindOption('o',opt.sDumpFile) && opt.sDumpFile.IsEmpty()==false ) {
     fOut = fopen( opt.sDumpFile.Str(), opt.isOutDOS ? "wt" : "wb" );
     if ( fOut==nil ) {
	 fprintf(stderr,"Unable to write file: %s\n",
		 opt.sDumpFile.Str());
	 return 4;
     }
 }

 // parsing nearly ends...
 if ( opt.caseKind>0 && opt.doViewAll ) {
     opt.caseKind = 0;
     if ( opt.isVerbose )
	 fprintf(fRepErr,"Warning: --view and -c options specified; ignoring -c (%s)\n",opt.sCase.Str());
 }
 // .

 UNI_STG_ADJUST;
 error = unicode_read();
 if ( error || opt.isVeryVerbose ) {
     char* pathStr( optUncode.UniThbFile().Str() );
     if ( error )
	 fprintf(fRepErr,"Unable to read Unicode-plus THB: %s, error: %d\n",
		 pathStr,
		 error);
     else
	 fprintf(fRepErr,"Did read Unicode-plus THB: %s\n",
		 pathStr);
 }
 if ( error ) return 2;

 thisError = unicode_leveling_latin1();
 DBGPRINT("DBG: unicode_leveling_latin1 returned %d\n",thisError);

 // go-ahead:
 error = go_cmd( arg, fOut, fRepErr, opt, 0 );
 DBGPRINT("DBG: go_cmd returned %d\n",error);

 return result = error;
}
////////////////////////////////////////////////////////////
int main (int argc, char** argv, char** envp)
{
 int error, result=0;

 gINIT;

 error = go( argv, envp, result );

 gEND;

 DBGPRINT_MIN("go: %d, result=%d\n",error,result);

 return result;
}

